package org.eclipse.jdt.internal.compiler.lookup;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import java.util.ArrayList;
import java.util.List;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.AbstractMethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.CaseStatement;
import org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable;
import org.eclipse.jdt.internal.compiler.ast.LambdaExpression;
import org.eclipse.jdt.internal.compiler.ast.LocalDeclaration;
import org.eclipse.jdt.internal.compiler.ast.ReturnStatement;
import org.eclipse.jdt.internal.compiler.ast.TypeDeclaration;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.flow.FlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.ClassScope;
import org.eclipse.jdt.internal.compiler.lookup.CompilationUnitScope;
import org.eclipse.jdt.internal.compiler.lookup.FieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.InvocationSite;
import org.eclipse.jdt.internal.compiler.lookup.LocalTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.LocalVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodScope;
import org.eclipse.jdt.internal.compiler.lookup.MethodVerifier;
import org.eclipse.jdt.internal.compiler.lookup.NestedTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.PackageBinding;
import org.eclipse.jdt.internal.compiler.lookup.ProblemBinding;
import org.eclipse.jdt.internal.compiler.lookup.ProblemFieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.ProblemReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.SourceTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.SyntheticArgumentBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.VariableBinding;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class BlockScope extends Scope {
  public LocalVariableBinding[] locals;
  
  public int localIndex;
  
  public int startIndex;
  
  public int offset;
  
  public int maxOffset;
  
  public org.eclipse.jdt.internal.compiler.lookup.BlockScope[] shiftScopes;
  
  public Scope[] subscopes = new Scope[1];
  
  public int subscopeCount = 0;
  
  public CaseStatement enclosingCase;
  
  public static final VariableBinding[] EmulationPathToImplicitThis = new VariableBinding[0];
  
  public static final VariableBinding[] NoEnclosingInstanceInConstructorCall = new VariableBinding[0];
  
  public static final VariableBinding[] NoEnclosingInstanceInStaticContext = new VariableBinding[0];
  
  public boolean insideTypeAnnotation = false;
  
  private List trackingVariables;
  
  public FlowInfo finallyInfo;
  
  public BlockScope(org.eclipse.jdt.internal.compiler.lookup.BlockScope parent) {
    this(parent, true);
  }
  
  public BlockScope(org.eclipse.jdt.internal.compiler.lookup.BlockScope parent, boolean addToParentScope) {
    this(1, parent);
    this.locals = new LocalVariableBinding[5];
    if (addToParentScope)
      parent.addSubscope(this); 
    this.startIndex = parent.localIndex;
  }
  
  public BlockScope(org.eclipse.jdt.internal.compiler.lookup.BlockScope parent, int variableCount) {
    this(1, parent);
    this.locals = new LocalVariableBinding[variableCount];
    parent.addSubscope(this);
    this.startIndex = parent.localIndex;
  }
  
  protected BlockScope(int kind, Scope parent) {
    super(kind, parent);
  }
  
  public final void addAnonymousType(TypeDeclaration anonymousType, ReferenceBinding superBinding) {
    ClassScope anonymousClassScope = new ClassScope(this, anonymousType);
    anonymousClassScope.buildAnonymousTypeBinding(
        enclosingSourceType(), 
        superBinding);
    MethodScope methodScope = methodScope();
    while (methodScope != null && methodScope.referenceContext instanceof LambdaExpression) {
      LambdaExpression lambda = (LambdaExpression)methodScope.referenceContext;
      if (!lambda.scope.isStatic && !lambda.scope.isConstructorCall)
        lambda.shouldCaptureInstance = true; 
      methodScope = methodScope.enclosingMethodScope();
    } 
  }
  
  public final void addLocalType(TypeDeclaration localType) {
    ClassScope localTypeScope = new ClassScope(this, localType);
    addSubscope((Scope)localTypeScope);
    localTypeScope.buildLocalTypeBinding(enclosingSourceType());
    MethodScope methodScope = methodScope();
    while (methodScope != null && methodScope.referenceContext instanceof LambdaExpression) {
      LambdaExpression lambda = (LambdaExpression)methodScope.referenceContext;
      if (!lambda.scope.isStatic && !lambda.scope.isConstructorCall)
        lambda.shouldCaptureInstance = true; 
      methodScope = methodScope.enclosingMethodScope();
    } 
  }
  
  public final void addLocalVariable(LocalVariableBinding binding) {
    checkAndSetModifiersForVariable(binding);
    if (this.localIndex == this.locals.length)
      System.arraycopy(
          this.locals, 
          0, 
          this.locals = new LocalVariableBinding[this.localIndex * 2], 
          0, 
          this.localIndex); 
    this.locals[this.localIndex++] = binding;
    binding.declaringScope = this;
    binding.id = (outerMostMethodScope()).analysisIndex++;
  }
  
  public void addSubscope(Scope childScope) {
    if (this.subscopeCount == this.subscopes.length)
      System.arraycopy(
          this.subscopes, 
          0, 
          this.subscopes = new Scope[this.subscopeCount * 2], 
          0, 
          this.subscopeCount); 
    this.subscopes[this.subscopeCount++] = childScope;
  }
  
  public final boolean allowBlankFinalFieldAssignment(FieldBinding binding) {
    if (TypeBinding.notEquals((TypeBinding)enclosingReceiverType(), (TypeBinding)binding.declaringClass))
      return false; 
    MethodScope methodScope = methodScope();
    if (methodScope.isStatic != binding.isStatic())
      return false; 
    if (methodScope.isLambdaScope())
      return false; 
    return !(!methodScope.isInsideInitializer() && 
      !((AbstractMethodDeclaration)methodScope.referenceContext).isInitializationMethod());
  }
  
  String basicToString(int tab) {
    String newLine = "\n";
    for (int i = tab; --i >= 0;)
      newLine = String.valueOf(newLine) + "\t"; 
    String s = String.valueOf(newLine) + "--- Block Scope ---";
    newLine = String.valueOf(newLine) + "\t";
    s = String.valueOf(s) + newLine + "locals:";
    for (int j = 0; j < this.localIndex; j++)
      s = String.valueOf(s) + newLine + "\t" + this.locals[j].toString(); 
    s = String.valueOf(s) + newLine + "startIndex = " + this.startIndex;
    return s;
  }
  
  private void checkAndSetModifiersForVariable(LocalVariableBinding varBinding) {
    int modifiers = varBinding.modifiers;
    if ((modifiers & 0x400000) != 0 && varBinding.declaration != null)
      problemReporter().duplicateModifierForVariable(varBinding.declaration, this instanceof MethodScope); 
    int realModifiers = modifiers & 0xFFFF;
    int unexpectedModifiers = -17;
    if ((realModifiers & unexpectedModifiers) != 0 && varBinding.declaration != null)
      problemReporter().illegalModifierForVariable(varBinding.declaration, this instanceof MethodScope); 
    varBinding.modifiers = modifiers;
  }
  
  void computeLocalVariablePositions(int ilocal, int initOffset, CodeStream codeStream) {
    this.offset = initOffset;
    this.maxOffset = initOffset;
    int maxLocals = this.localIndex;
    boolean hasMoreVariables = (ilocal < maxLocals);
    int iscope = 0, maxScopes = this.subscopeCount;
    boolean hasMoreScopes = (maxScopes > 0);
    while (hasMoreVariables || hasMoreScopes) {
      if (hasMoreScopes && (
        !hasMoreVariables || this.subscopes[iscope].startIndex() <= ilocal)) {
        if (this.subscopes[iscope] instanceof org.eclipse.jdt.internal.compiler.lookup.BlockScope) {
          org.eclipse.jdt.internal.compiler.lookup.BlockScope subscope = (org.eclipse.jdt.internal.compiler.lookup.BlockScope)this.subscopes[iscope];
          int subOffset = (subscope.shiftScopes == null) ? this.offset : subscope.maxShiftedOffset();
          subscope.computeLocalVariablePositions(0, subOffset, codeStream);
          if (subscope.maxOffset > this.maxOffset)
            this.maxOffset = subscope.maxOffset; 
        } 
        hasMoreScopes = (++iscope < maxScopes);
        continue;
      } 
      LocalVariableBinding local = this.locals[ilocal];
      boolean generateCurrentLocalVar = (local.useFlag > 0 && local.constant() == Constant.NotAConstant);
      if (local.useFlag == 0 && 
        local.declaration != null && (
        local.declaration.bits & 0x40000000) != 0)
        if (local.isCatchParameter()) {
          problemReporter().unusedExceptionParameter(local.declaration);
        } else {
          problemReporter().unusedLocalVariable(local.declaration);
        }  
      if (!generateCurrentLocalVar && 
        local.declaration != null && (compilerOptions()).preserveAllLocalVariables) {
        generateCurrentLocalVar = true;
        if (local.useFlag == 0)
          local.useFlag = 1; 
      } 
      if (generateCurrentLocalVar) {
        if (local.declaration != null)
          codeStream.record(local); 
        local.resolvedPosition = this.offset;
        if (TypeBinding.equalsEquals(local.type, (TypeBinding)TypeBinding.LONG) || TypeBinding.equalsEquals(local.type, (TypeBinding)TypeBinding.DOUBLE)) {
          this.offset += 2;
        } else {
          this.offset++;
        } 
        if (this.offset > 65535)
          problemReporter().noMoreAvailableSpaceForLocal(
              local, 
              (local.declaration == null) ? (ASTNode)(methodScope()).referenceContext : (ASTNode)local.declaration); 
      } else {
        local.resolvedPosition = -1;
      } 
      hasMoreVariables = (++ilocal < maxLocals);
    } 
    if (this.offset > this.maxOffset)
      this.maxOffset = this.offset; 
  }
  
  public void emulateOuterAccess(LocalVariableBinding outerLocalVariable) {
    org.eclipse.jdt.internal.compiler.lookup.BlockScope outerVariableScope = outerLocalVariable.declaringScope;
    if (outerVariableScope == null)
      return; 
    int depth = 0;
    Scope scope = this;
    while (outerVariableScope != scope) {
      switch (scope.kind) {
        case 3:
          depth++;
          break;
        case 2:
          if (scope.isLambdaScope()) {
            LambdaExpression lambdaExpression = (LambdaExpression)scope.referenceContext();
            lambdaExpression.addSyntheticArgument(outerLocalVariable);
          } 
          break;
      } 
      scope = scope.parent;
    } 
    if (depth == 0)
      return; 
    MethodScope currentMethodScope = methodScope();
    if (outerVariableScope.methodScope() != currentMethodScope) {
      NestedTypeBinding currentType = (NestedTypeBinding)enclosingSourceType();
      if (!currentType.isLocalType())
        return; 
      if (!currentMethodScope.isInsideInitializerOrConstructor()) {
        currentType.addSyntheticArgumentAndField(outerLocalVariable);
      } else {
        currentType.addSyntheticArgument(outerLocalVariable);
      } 
    } 
  }
  
  public final ReferenceBinding findLocalType(char[] name) {
    long compliance = (compilerOptions()).complianceLevel;
    for (int i = this.subscopeCount - 1; i >= 0; i--) {
      if (this.subscopes[i] instanceof ClassScope) {
        LocalTypeBinding sourceType = (LocalTypeBinding)((ClassScope)this.subscopes[i]).referenceContext.binding;
        if (compliance < 3145728L || sourceType.enclosingCase == null || 
          isInsideCase(sourceType.enclosingCase))
          if (CharOperation.equals(sourceType.sourceName(), name))
            return (ReferenceBinding)sourceType;  
      } 
    } 
    return null;
  }
  
  public LocalDeclaration[] findLocalVariableDeclarations(int position) {
    int ilocal = 0, maxLocals = this.localIndex;
    boolean hasMoreVariables = (maxLocals > 0);
    LocalDeclaration[] localDeclarations = null;
    int declPtr = 0;
    int iscope = 0, maxScopes = this.subscopeCount;
    boolean hasMoreScopes = (maxScopes > 0);
    while (hasMoreVariables || hasMoreScopes) {
      if (hasMoreScopes && (
        !hasMoreVariables || this.subscopes[iscope].startIndex() <= ilocal)) {
        Scope subscope = this.subscopes[iscope];
        if (subscope.kind == 1) {
          localDeclarations = ((org.eclipse.jdt.internal.compiler.lookup.BlockScope)subscope).findLocalVariableDeclarations(position);
          if (localDeclarations != null)
            return localDeclarations; 
        } 
        hasMoreScopes = (++iscope < maxScopes);
        continue;
      } 
      LocalVariableBinding local = this.locals[ilocal];
      if (local != null) {
        LocalDeclaration localDecl = local.declaration;
        if (localDecl != null)
          if (localDecl.declarationSourceStart <= position) {
            if (position <= localDecl.declarationSourceEnd) {
              if (localDeclarations == null)
                localDeclarations = new LocalDeclaration[maxLocals]; 
              localDeclarations[declPtr++] = localDecl;
            } 
          } else {
            return localDeclarations;
          }  
      } 
      hasMoreVariables = (++ilocal < maxLocals);
      if (!hasMoreVariables && localDeclarations != null)
        return localDeclarations; 
    } 
    return null;
  }
  
  public LocalVariableBinding findVariable(char[] variableName) {
    int varLength = variableName.length;
    for (int i = this.localIndex - 1; i >= 0; i--) {
      LocalVariableBinding local;
      char[] localName;
      if ((localName = (local = this.locals[i]).name).length == varLength && CharOperation.equals(localName, variableName))
        return local; 
    } 
    return null;
  }
  
  public Binding getBinding(char[][] compoundName, int mask, InvocationSite invocationSite, boolean needResolve) {
    ReferenceBinding referenceBinding1;
    Binding binding = getBinding(compoundName[0], mask | 0x4 | 0x10, invocationSite, needResolve);
    invocationSite.setFieldIndex(1);
    if (binding instanceof VariableBinding)
      return binding; 
    CompilationUnitScope unitScope = compilationUnitScope();
    unitScope.recordQualifiedReference(compoundName);
    if (!binding.isValidBinding())
      return binding; 
    int length = compoundName.length;
    int currentIndex = 1;
    if (binding instanceof PackageBinding) {
      PackageBinding packageBinding = (PackageBinding)binding;
      while (currentIndex < length) {
        unitScope.recordReference(packageBinding.compoundName, compoundName[currentIndex]);
        binding = packageBinding.getTypeOrPackage(compoundName[currentIndex++]);
        invocationSite.setFieldIndex(currentIndex);
        if (binding == null) {
          if (currentIndex == length)
            return (Binding)new ProblemReferenceBinding(
                CharOperation.subarray(compoundName, 0, currentIndex), 
                null, 
                1); 
          return (Binding)new ProblemBinding(
              CharOperation.subarray(compoundName, 0, currentIndex), 
              1);
        } 
        if (binding instanceof ReferenceBinding) {
          if (!binding.isValidBinding())
            return (Binding)new ProblemReferenceBinding(
                CharOperation.subarray(compoundName, 0, currentIndex), 
                (ReferenceBinding)((ReferenceBinding)binding).closestMatch(), 
                binding.problemId()); 
          if (!((ReferenceBinding)binding).canBeSeenBy(this))
            return (Binding)new ProblemReferenceBinding(
                CharOperation.subarray(compoundName, 0, currentIndex), 
                (ReferenceBinding)binding, 
                2); 
          // Byte code: goto -> 276
        } 
        packageBinding = (PackageBinding)binding;
      } 
      return (Binding)new ProblemReferenceBinding(
          CharOperation.subarray(compoundName, 0, currentIndex), 
          null, 
          1);
    } 
    ReferenceBinding referenceBinding = (ReferenceBinding)binding;
    TypeBinding typeBinding = environment().convertToRawType((TypeBinding)referenceBinding, false);
    if (invocationSite instanceof ASTNode) {
      ASTNode invocationNode = (ASTNode)invocationSite;
      if (invocationNode.isTypeUseDeprecated((TypeBinding)referenceBinding, this))
        problemReporter().deprecatedType((TypeBinding)referenceBinding, invocationNode); 
    } 
    Binding problemFieldBinding = null;
    while (currentIndex < length) {
      ProblemFieldBinding problemFieldBinding1;
      referenceBinding = (ReferenceBinding)typeBinding;
      char[] nextName = compoundName[currentIndex++];
      invocationSite.setFieldIndex(currentIndex);
      invocationSite.setActualReceiverType(referenceBinding);
      FieldBinding fieldBinding;
      if ((mask & 0x1) != 0 && (fieldBinding = findField((TypeBinding)referenceBinding, nextName, invocationSite, true)) != null) {
        if (fieldBinding.isValidBinding())
          break; 
        problemFieldBinding1 = new ProblemFieldBinding(
            ((ProblemFieldBinding)fieldBinding).closestMatch, 
            ((ProblemFieldBinding)fieldBinding).declaringClass, 
            CharOperation.concatWith(CharOperation.subarray(compoundName, 0, currentIndex), '.'), 
            fieldBinding.problemId());
        if (fieldBinding.problemId() != 2)
          return (Binding)problemFieldBinding1; 
      } 
      if ((referenceBinding1 = findMemberType(nextName, referenceBinding)) == null) {
        if (problemFieldBinding1 != null)
          return (Binding)problemFieldBinding1; 
        if ((mask & 0x1) != 0)
          return (Binding)new ProblemFieldBinding(
              null, 
              referenceBinding, 
              nextName, 
              1); 
        if ((mask & 0x3) != 0)
          return (Binding)new ProblemBinding(
              CharOperation.subarray(compoundName, 0, currentIndex), 
              referenceBinding, 
              1); 
        return (Binding)new ProblemReferenceBinding(
            CharOperation.subarray(compoundName, 0, currentIndex), 
            referenceBinding, 
            1);
      } 
      if (!referenceBinding1.isValidBinding()) {
        if (problemFieldBinding1 != null)
          return (Binding)problemFieldBinding1; 
        return (Binding)new ProblemReferenceBinding(
            CharOperation.subarray(compoundName, 0, currentIndex), 
            (ReferenceBinding)referenceBinding1.closestMatch(), 
            referenceBinding1.problemId());
      } 
      if (invocationSite instanceof ASTNode) {
        referenceBinding = referenceBinding1;
        ASTNode invocationNode = (ASTNode)invocationSite;
        if (invocationNode.isTypeUseDeprecated((TypeBinding)referenceBinding, this))
          problemReporter().deprecatedType((TypeBinding)referenceBinding, invocationNode); 
      } 
    } 
    if ((mask & 0x1) != 0 && referenceBinding1 instanceof FieldBinding) {
      FieldBinding field = (FieldBinding)referenceBinding1;
      if (!field.isStatic())
        return (Binding)new ProblemFieldBinding(
            field, 
            field.declaringClass, 
            CharOperation.concatWith(CharOperation.subarray(compoundName, 0, currentIndex), '.'), 
            7); 
      return (Binding)referenceBinding1;
    } 
    if ((mask & 0x4) != 0 && referenceBinding1 instanceof ReferenceBinding)
      return (Binding)referenceBinding1; 
    return (Binding)new ProblemBinding(
        CharOperation.subarray(compoundName, 0, currentIndex), 
        1);
  }
  
  public final Binding getBinding(char[][] compoundName, InvocationSite invocationSite) {
    ReferenceBinding referenceBinding;
    FieldBinding fieldBinding;
    int currentIndex = 0;
    int length = compoundName.length;
    Binding binding = 
      getBinding(
        compoundName[currentIndex++], 
        23, 
        invocationSite, 
        true);
    if (!binding.isValidBinding())
      return binding; 
    if (binding instanceof PackageBinding)
      while (true) {
        if (currentIndex >= length)
          return binding; 
        PackageBinding packageBinding = (PackageBinding)binding;
        binding = packageBinding.getTypeOrPackage(compoundName[currentIndex++]);
        if (binding == null) {
          if (currentIndex == length)
            return (Binding)new ProblemReferenceBinding(CharOperation.subarray(compoundName, 0, currentIndex), null, 1); 
          return (Binding)new ProblemBinding(CharOperation.subarray(compoundName, 0, currentIndex), 1);
        } 
        if (binding instanceof ReferenceBinding) {
          if (!binding.isValidBinding())
            return (Binding)new ProblemReferenceBinding(CharOperation.subarray(compoundName, 0, currentIndex), (ReferenceBinding)((ReferenceBinding)binding).closestMatch(), binding.problemId()); 
          if (!((ReferenceBinding)binding).canBeSeenBy(this))
            return (Binding)new ProblemReferenceBinding(CharOperation.subarray(compoundName, 0, currentIndex), (ReferenceBinding)binding, 2); 
          break;
        } 
      }  
    if (binding instanceof ReferenceBinding)
      while (true) {
        if (currentIndex >= length)
          return binding; 
        ReferenceBinding typeBinding = (ReferenceBinding)binding;
        char[] nextName = compoundName[currentIndex++];
        TypeBinding receiverType = typeBinding.capture(this, invocationSite.sourceStart(), invocationSite.sourceEnd());
        FieldBinding fieldBinding1;
        if ((fieldBinding1 = findField(receiverType, nextName, invocationSite, true)) != null) {
          if (!fieldBinding1.isValidBinding())
            return (Binding)new ProblemFieldBinding(fieldBinding1, fieldBinding1.declaringClass, CharOperation.concatWith(CharOperation.subarray(compoundName, 0, currentIndex), '.'), fieldBinding1.problemId()); 
          if (!fieldBinding1.isStatic())
            return (Binding)new ProblemFieldBinding(fieldBinding1, fieldBinding1.declaringClass, CharOperation.concatWith(CharOperation.subarray(compoundName, 0, currentIndex), '.'), 7); 
          break;
        } 
        if ((referenceBinding = findMemberType(nextName, typeBinding)) == null)
          return (Binding)new ProblemBinding(CharOperation.subarray(compoundName, 0, currentIndex), typeBinding, 1); 
        if (!referenceBinding.isValidBinding())
          return (Binding)new ProblemReferenceBinding(CharOperation.subarray(compoundName, 0, currentIndex), (ReferenceBinding)referenceBinding.closestMatch(), referenceBinding.problemId()); 
      }  
    VariableBinding variableBinding = (VariableBinding)referenceBinding;
    while (currentIndex < length) {
      TypeBinding typeBinding = variableBinding.type;
      if (typeBinding == null)
        return (Binding)new ProblemFieldBinding(
            null, 
            null, 
            CharOperation.concatWith(CharOperation.subarray(compoundName, 0, currentIndex), '.'), 
            1); 
      TypeBinding receiverType = typeBinding.capture(this, invocationSite.sourceStart(), invocationSite.sourceEnd());
      fieldBinding = findField(receiverType, compoundName[currentIndex++], invocationSite, true);
      if (fieldBinding == null)
        return (Binding)new ProblemFieldBinding(
            null, 
            (receiverType instanceof ReferenceBinding) ? (ReferenceBinding)receiverType : null, 
            CharOperation.concatWith(CharOperation.subarray(compoundName, 0, currentIndex), '.'), 
            1); 
      if (!fieldBinding.isValidBinding())
        return (Binding)fieldBinding; 
    } 
    return (Binding)fieldBinding;
  }
  
  public VariableBinding[] getEmulationPath(LocalVariableBinding outerLocalVariable) {
    MethodScope currentMethodScope = methodScope();
    SourceTypeBinding sourceType = currentMethodScope.enclosingSourceType();
    org.eclipse.jdt.internal.compiler.lookup.BlockScope variableScope = outerLocalVariable.declaringScope;
    if (variableScope == null || currentMethodScope == variableScope.methodScope())
      return new VariableBinding[] { (VariableBinding)outerLocalVariable }; 
    LambdaExpression lambda = (LambdaExpression)currentMethodScope.referenceContext;
    SyntheticArgumentBinding syntheticArgument;
    if (currentMethodScope.isLambdaScope() && (syntheticArgument = lambda.getSyntheticArgument(outerLocalVariable)) != null)
      return new VariableBinding[] { (VariableBinding)syntheticArgument }; 
    if (currentMethodScope.isInsideInitializerOrConstructor() && 
      sourceType.isNestedType()) {
      SyntheticArgumentBinding syntheticArg;
      if ((syntheticArg = ((NestedTypeBinding)sourceType).getSyntheticArgument(outerLocalVariable)) != null)
        return new VariableBinding[] { (VariableBinding)syntheticArg }; 
    } 
    if (!currentMethodScope.isStatic) {
      FieldBinding syntheticField;
      if ((syntheticField = sourceType.getSyntheticField(outerLocalVariable)) != null)
        return new VariableBinding[] { (VariableBinding)syntheticField }; 
    } 
    return null;
  }
  
  public Object[] getEmulationPath(ReferenceBinding targetEnclosingType, boolean onlyExactMatch, boolean denyEnclosingArgInConstructorCall) {
    MethodScope currentMethodScope = methodScope();
    SourceTypeBinding sourceType = currentMethodScope.enclosingSourceType();
    if (!currentMethodScope.isStatic && !currentMethodScope.isConstructorCall && (
      TypeBinding.equalsEquals((TypeBinding)sourceType, (TypeBinding)targetEnclosingType) || (!onlyExactMatch && sourceType.findSuperTypeOriginatingFrom((TypeBinding)targetEnclosingType) != null)))
      return (Object[])EmulationPathToImplicitThis; 
    if (!sourceType.isNestedType() || sourceType.isStatic()) {
      if (currentMethodScope.isConstructorCall)
        return (Object[])NoEnclosingInstanceInConstructorCall; 
      if (currentMethodScope.isStatic)
        return (Object[])NoEnclosingInstanceInStaticContext; 
      return null;
    } 
    boolean insideConstructor = currentMethodScope.isInsideInitializerOrConstructor();
    if (insideConstructor) {
      SyntheticArgumentBinding syntheticArg;
      if ((syntheticArg = ((NestedTypeBinding)sourceType).getSyntheticArgument(targetEnclosingType, onlyExactMatch, currentMethodScope.isConstructorCall)) != null) {
        boolean isAnonymousAndHasEnclosing = (sourceType.isAnonymousType() && 
          sourceType.scope.referenceContext.allocation.enclosingInstance != null);
        if (denyEnclosingArgInConstructorCall && 
          currentMethodScope.isConstructorCall && 
          !isAnonymousAndHasEnclosing && (
          TypeBinding.equalsEquals((TypeBinding)sourceType, (TypeBinding)targetEnclosingType) || (!onlyExactMatch && sourceType.findSuperTypeOriginatingFrom((TypeBinding)targetEnclosingType) != null)))
          return (Object[])NoEnclosingInstanceInConstructorCall; 
        return new Object[] { syntheticArg };
      } 
    } 
    if (currentMethodScope.isStatic)
      return (Object[])NoEnclosingInstanceInStaticContext; 
    if (sourceType.isAnonymousType()) {
      ReferenceBinding enclosingType = sourceType.enclosingType();
      if (enclosingType.isNestedType()) {
        NestedTypeBinding nestedEnclosingType = (NestedTypeBinding)enclosingType;
        SyntheticArgumentBinding enclosingArgument = nestedEnclosingType.getSyntheticArgument(nestedEnclosingType.enclosingType(), onlyExactMatch, currentMethodScope.isConstructorCall);
        if (enclosingArgument != null) {
          FieldBinding fieldBinding = sourceType.getSyntheticField((LocalVariableBinding)enclosingArgument);
          if (fieldBinding != null && (
            TypeBinding.equalsEquals(fieldBinding.type, (TypeBinding)targetEnclosingType) || (!onlyExactMatch && ((ReferenceBinding)fieldBinding.type).findSuperTypeOriginatingFrom((TypeBinding)targetEnclosingType) != null)))
            return new Object[] { fieldBinding }; 
        } 
      } 
    } 
    FieldBinding syntheticField = sourceType.getSyntheticField(targetEnclosingType, onlyExactMatch);
    if (syntheticField != null) {
      if (currentMethodScope.isConstructorCall)
        return (Object[])NoEnclosingInstanceInConstructorCall; 
      return new Object[] { syntheticField };
    } 
    Object[] path = new Object[2];
    ReferenceBinding currentType = sourceType.enclosingType();
    if (insideConstructor) {
      path[0] = ((NestedTypeBinding)sourceType).getSyntheticArgument(currentType, onlyExactMatch, currentMethodScope.isConstructorCall);
    } else {
      if (currentMethodScope.isConstructorCall)
        return (Object[])NoEnclosingInstanceInConstructorCall; 
      path[0] = sourceType.getSyntheticField(currentType, onlyExactMatch);
    } 
    if (path[0] != null) {
      int count = 1;
      ReferenceBinding currentEnclosingType;
      while ((currentEnclosingType = currentType.enclosingType()) != null) {
        if (TypeBinding.equalsEquals((TypeBinding)currentType, (TypeBinding)targetEnclosingType) || (
          !onlyExactMatch && currentType.findSuperTypeOriginatingFrom((TypeBinding)targetEnclosingType) != null))
          break; 
        if (currentMethodScope != null) {
          currentMethodScope = currentMethodScope.enclosingMethodScope();
          if (currentMethodScope != null && currentMethodScope.isConstructorCall)
            return (Object[])NoEnclosingInstanceInConstructorCall; 
          if (currentMethodScope != null && currentMethodScope.isStatic)
            return (Object[])NoEnclosingInstanceInStaticContext; 
        } 
        syntheticField = ((NestedTypeBinding)currentType).getSyntheticField(currentEnclosingType, onlyExactMatch);
        if (syntheticField == null)
          break; 
        if (count == path.length)
          System.arraycopy(path, 0, path = new Object[count + 1], 0, count); 
        path[count++] = ((SourceTypeBinding)syntheticField.declaringClass).addSyntheticMethod(syntheticField, true, false);
        currentType = currentEnclosingType;
      } 
      if (TypeBinding.equalsEquals((TypeBinding)currentType, (TypeBinding)targetEnclosingType) || (
        !onlyExactMatch && currentType.findSuperTypeOriginatingFrom((TypeBinding)targetEnclosingType) != null))
        return path; 
    } 
    return null;
  }
  
  public final boolean isDuplicateLocalVariable(char[] name) {
    org.eclipse.jdt.internal.compiler.lookup.BlockScope current = this;
    while (true) {
      for (int i = 0; i < this.localIndex; i++) {
        if (CharOperation.equals(name, (current.locals[i]).name))
          return true; 
      } 
      if (current.kind != 1)
        return false; 
      current = (org.eclipse.jdt.internal.compiler.lookup.BlockScope)current.parent;
    } 
  }
  
  public int maxShiftedOffset() {
    int max = -1;
    if (this.shiftScopes != null)
      for (int i = 0, length = this.shiftScopes.length; i < length; i++) {
        if (this.shiftScopes[i] != null) {
          int subMaxOffset = (this.shiftScopes[i]).maxOffset;
          if (subMaxOffset > max)
            max = subMaxOffset; 
        } 
      }  
    return max;
  }
  
  public final boolean needBlankFinalFieldInitializationCheck(FieldBinding binding) {
    boolean isStatic = binding.isStatic();
    ReferenceBinding fieldDeclaringClass = binding.declaringClass;
    MethodScope methodScope = namedMethodScope();
    while (methodScope != null) {
      if (methodScope.isStatic != isStatic)
        return false; 
      if (!methodScope.isInsideInitializer() && 
        !((AbstractMethodDeclaration)methodScope.referenceContext).isInitializationMethod())
        return false; 
      ReferenceBinding enclosingType = methodScope.enclosingReceiverType();
      if (TypeBinding.equalsEquals((TypeBinding)enclosingType, (TypeBinding)fieldDeclaringClass))
        return true; 
      if (!enclosingType.erasure().isAnonymousType())
        return false; 
      methodScope = methodScope.enclosingMethodScope().namedMethodScope();
    } 
    return false;
  }
  
  public ProblemReporter problemReporter() {
    return methodScope().problemReporter();
  }
  
  public void propagateInnerEmulation(ReferenceBinding targetType, boolean isEnclosingInstanceSupplied) {
    SyntheticArgumentBinding[] syntheticArguments;
    if ((syntheticArguments = targetType.syntheticOuterLocalVariables()) != null)
      for (int i = 0, max = syntheticArguments.length; i < max; i++) {
        SyntheticArgumentBinding syntheticArg = syntheticArguments[i];
        if (!isEnclosingInstanceSupplied || 
          !TypeBinding.equalsEquals(syntheticArg.type, (TypeBinding)targetType.enclosingType()))
          emulateOuterAccess(syntheticArg.actualOuterLocalVariable); 
      }  
  }
  
  public TypeDeclaration referenceType() {
    return methodScope().referenceType();
  }
  
  public int scopeIndex() {
    if (this instanceof MethodScope)
      return -1; 
    org.eclipse.jdt.internal.compiler.lookup.BlockScope parentScope = (org.eclipse.jdt.internal.compiler.lookup.BlockScope)this.parent;
    Scope[] parentSubscopes = parentScope.subscopes;
    for (int i = 0, max = parentScope.subscopeCount; i < max; i++) {
      if (parentSubscopes[i] == this)
        return i; 
    } 
    return -1;
  }
  
  int startIndex() {
    return this.startIndex;
  }
  
  public String toString() {
    return toString(0);
  }
  
  public String toString(int tab) {
    String s = basicToString(tab);
    for (int i = 0; i < this.subscopeCount; i++) {
      if (this.subscopes[i] instanceof org.eclipse.jdt.internal.compiler.lookup.BlockScope)
        s = String.valueOf(s) + ((org.eclipse.jdt.internal.compiler.lookup.BlockScope)this.subscopes[i]).toString(tab + 1) + "\n"; 
    } 
    return s;
  }
  
  public int registerTrackingVariable(FakedTrackingVariable fakedTrackingVariable) {
    if (this.trackingVariables == null)
      this.trackingVariables = new ArrayList(3); 
    this.trackingVariables.add(fakedTrackingVariable);
    MethodScope outerMethodScope = outerMostMethodScope();
    return outerMethodScope.analysisIndex++;
  }
  
  public void removeTrackingVar(FakedTrackingVariable trackingVariable) {
    if (trackingVariable.innerTracker != null) {
      trackingVariable.innerTracker.withdraw();
      trackingVariable.innerTracker = null;
    } 
    if (this.trackingVariables != null && 
      this.trackingVariables.remove(trackingVariable))
      return; 
    if (this.parent instanceof org.eclipse.jdt.internal.compiler.lookup.BlockScope)
      ((org.eclipse.jdt.internal.compiler.lookup.BlockScope)this.parent).removeTrackingVar(trackingVariable); 
  }
  
  public void pruneWrapperTrackingVar(FakedTrackingVariable trackingVariable) {
    this.trackingVariables.remove(trackingVariable);
  }
  
  public void checkUnclosedCloseables(FlowInfo flowInfo, FlowContext flowContext, ASTNode location, org.eclipse.jdt.internal.compiler.lookup.BlockScope locationScope) {
    if (!(compilerOptions()).analyseResourceLeaks)
      return; 
    if (this.trackingVariables == null) {
      if (location != null && this.parent instanceof org.eclipse.jdt.internal.compiler.lookup.BlockScope)
        ((org.eclipse.jdt.internal.compiler.lookup.BlockScope)this.parent).checkUnclosedCloseables(flowInfo, flowContext, location, locationScope); 
      return;
    } 
    if (location != null && flowInfo.reachMode() != 0)
      return; 
    FakedTrackingVariable returnVar = (location instanceof ReturnStatement) ? 
      FakedTrackingVariable.getCloseTrackingVariable(((ReturnStatement)location).expression, flowInfo, flowContext) : null;
    FakedTrackingVariable.IteratorForReporting<FakedTrackingVariable> iteratorForReporting = new FakedTrackingVariable.IteratorForReporting(this.trackingVariables, this, (location != null));
    while (iteratorForReporting.hasNext()) {
      FakedTrackingVariable trackingVar = iteratorForReporting.next();
      if (returnVar != null && trackingVar.isResourceBeingReturned(returnVar))
        continue; 
      if (location != null && trackingVar.hasDefinitelyNoResource(flowInfo))
        continue; 
      if (location != null && flowContext != null && flowContext.recordExitAgainstResource(this, flowInfo, trackingVar, location))
        continue; 
      int status = trackingVar.findMostSpecificStatus(flowInfo, this, locationScope);
      if (status == 2) {
        reportResourceLeak(trackingVar, location, status);
        continue;
      } 
      if (location == null)
        if (trackingVar.reportRecordedErrors(this, status, (flowInfo.reachMode() != 0)))
          continue;  
      if (status == 16) {
        reportResourceLeak(trackingVar, location, status);
        continue;
      } 
      if (status == 4)
        if ((environment()).globalOptions.complianceLevel >= 3342336L)
          trackingVar.reportExplicitClosing(problemReporter());  
    } 
    if (location == null) {
      for (int i = 0; i < this.localIndex; i++)
        (this.locals[i]).closeTracker = null; 
      this.trackingVariables = null;
    } 
  }
  
  private void reportResourceLeak(FakedTrackingVariable trackingVar, ASTNode location, int nullStatus) {
    if (location != null) {
      trackingVar.recordErrorLocation(location, nullStatus);
    } else {
      trackingVar.reportError(problemReporter(), null, nullStatus);
    } 
  }
  
  public void correlateTrackingVarsIfElse(FlowInfo thenFlowInfo, FlowInfo elseFlowInfo) {
    if (this.trackingVariables != null) {
      int trackVarCount = this.trackingVariables.size();
      for (int i = 0; i < trackVarCount; i++) {
        FakedTrackingVariable trackingVar = this.trackingVariables.get(i);
        if (trackingVar.originalBinding != null)
          if (thenFlowInfo.isDefinitelyNonNull(trackingVar.binding) && 
            elseFlowInfo.isDefinitelyNull(trackingVar.originalBinding)) {
            elseFlowInfo.markAsDefinitelyNonNull(trackingVar.binding);
          } else if (elseFlowInfo.isDefinitelyNonNull(trackingVar.binding) && 
            thenFlowInfo.isDefinitelyNull(trackingVar.originalBinding)) {
            thenFlowInfo.markAsDefinitelyNonNull(trackingVar.binding);
          } else if (thenFlowInfo != FlowInfo.DEAD_END && elseFlowInfo != FlowInfo.DEAD_END) {
            for (int j = i + 1; j < trackVarCount; j++) {
              FakedTrackingVariable var2 = this.trackingVariables.get(j);
              if (trackingVar.originalBinding == var2.originalBinding) {
                int newStatus;
                boolean var1SeenInThen = thenFlowInfo.hasNullInfoFor(trackingVar.binding);
                boolean var1SeenInElse = elseFlowInfo.hasNullInfoFor(trackingVar.binding);
                boolean var2SeenInThen = thenFlowInfo.hasNullInfoFor(var2.binding);
                boolean var2SeenInElse = elseFlowInfo.hasNullInfoFor(var2.binding);
                if (!var1SeenInThen && var1SeenInElse && var2SeenInThen && !var2SeenInElse) {
                  newStatus = FlowInfo.mergeNullStatus(thenFlowInfo.nullStatus(var2.binding), elseFlowInfo.nullStatus(trackingVar.binding));
                } else if (var1SeenInThen && !var1SeenInElse && !var2SeenInThen && var2SeenInElse) {
                  newStatus = FlowInfo.mergeNullStatus(thenFlowInfo.nullStatus(trackingVar.binding), elseFlowInfo.nullStatus(var2.binding));
                } else {
                  continue;
                } 
                thenFlowInfo.markNullStatus(trackingVar.binding, newStatus);
                elseFlowInfo.markNullStatus(trackingVar.binding, newStatus);
                trackingVar.originalBinding.closeTracker = trackingVar;
                thenFlowInfo.markNullStatus(var2.binding, 4);
                elseFlowInfo.markNullStatus(var2.binding, 4);
              } 
              continue;
            } 
          }  
      } 
    } 
    if (this.parent instanceof org.eclipse.jdt.internal.compiler.lookup.BlockScope)
      ((org.eclipse.jdt.internal.compiler.lookup.BlockScope)this.parent).correlateTrackingVarsIfElse(thenFlowInfo, elseFlowInfo); 
  }
  
  public void checkAppropriateMethodAgainstSupers(char[] selector, MethodBinding compileTimeMethod, TypeBinding[] parameters, InvocationSite site) {
    ReferenceBinding enclosingType = enclosingReceiverType();
    MethodBinding otherMethod = getMethod((TypeBinding)enclosingType.superclass(), selector, parameters, site);
    if (checkAppropriate(compileTimeMethod, otherMethod, site)) {
      ReferenceBinding[] superInterfaces = enclosingType.superInterfaces();
      if (superInterfaces != null)
        for (int i = 0; i < superInterfaces.length; i++) {
          otherMethod = getMethod((TypeBinding)superInterfaces[i], selector, parameters, site);
          if (!checkAppropriate(compileTimeMethod, otherMethod, site))
            break; 
        }  
    } 
  }
  
  private boolean checkAppropriate(MethodBinding compileTimeDeclaration, MethodBinding otherMethod, InvocationSite location) {
    if (otherMethod == null || !otherMethod.isValidBinding() || otherMethod == compileTimeDeclaration)
      return true; 
    if (MethodVerifier.doesMethodOverride(otherMethod, compileTimeDeclaration, environment())) {
      problemReporter().illegalSuperCallBypassingOverride(location, compileTimeDeclaration, otherMethod.declaringClass);
      return false;
    } 
    return true;
  }
  
  public boolean hasDefaultNullnessFor(int location) {
    return this.parent.hasDefaultNullnessFor(location);
  }
}
